use std::io;
use std::fs::remove_dir_all;
use std::path::Path;

use log::{debug, trace};

use errors::{BinError, BuildError, InstallError};
use manifest::Manifest;
use pkgfile::PKGFile;

use crate::tmpfs::TempDir;
use crate::types::fetch::TryFetch;
use crate::util::fs::copy_recursively;

pub mod errors;

#[derive(Debug)]
pub enum InstallType {
    Build,
    Bin,
}

#[derive(Debug)]
pub struct PackageInstaller {
    manifest: Manifest,
    pkgfile: PKGFile,
    install_type: InstallType,
}

impl PackageInstaller {
    pub fn new(m: Manifest, p: PKGFile, i: InstallType) -> PackageInstaller {
        PackageInstaller {
            manifest: m,
            pkgfile: p,
            install_type: i,
        }
    }

    fn extract_to(&self, path: String) -> Result<(), BinError> {
        if !self.pkgfile.has_data() {
            return Err(BinError::UnpackError("package has no data".to_string()));
        }
        let path = std::path::Path::new(&path);
        if path.exists() {
            trace!("cache already exists..");
            debug!("removing cache dir...");
            match remove_dir_all(path) {
                Ok(_) => debug!("removed cache directory"),
                Err(e) => return Err(BinError::UnpackError(format!("unable to remove directory ({}): {}", path.to_str().unwrap_or("unknown"), e.to_string())))
            }
        }
        tar::Archive::new(self.pkgfile.data.as_slice())
            .unpack(&path)
            .map_err(|e| BinError::UnpackError(e.to_string()))
    }

    fn bin(&self) -> Result<(), BinError> {
        let mut tmpdir = TempDir::default();
        tmpdir.push(&self.manifest.package.name);
        trace!("extracting package into: {}", tmpdir.to_string());
        if let Err(e) = self.extract_to(tmpdir.to_string()) {
            return Err(e);
        }
        debug!("extracted package in: {}", tmpdir.to_string());
        match self.apply_overlay() {
            Ok(_) => Ok(()),
            Err(e) => Err(BinError::IOError(e))
        }
    }

    fn apply_overlay(&self) -> Result<(), io::Error> {
        let mut tmpdir = TempDir::default();
        tmpdir.push(&self.manifest.package.name);
        tmpdir.push(&self.manifest.bin.clone().unwrap().root);
        if let Err(e) = copy_recursively(&tmpdir.path(), Path::new("/")) {
            return Err(e);
        }
        Ok(())
    }

    fn build(&self) -> Result<(), BuildError> {
        if let None = self.manifest.build.clone() {
            return Err(BuildError::InvalidManifest(String::from("No build manifest")));
        }
        let build_manifest = self.manifest.build.clone().unwrap();
        // TODO: Check dependencies

        Ok(())
    }

    pub fn install(&mut self) -> Result<(), InstallError> {
        match self.install_type {
            InstallType::Bin => {
                if let None = self.manifest.bin {
                    self.install_type = InstallType::Build;
                    if let None = self.manifest.build {
                        return Err(InstallError::InvalidManifest(String::from("no bin or build manifest")));
                    }
                }
            }
            InstallType::Build => {
                if let None = self.manifest.build {
                    self.install_type = InstallType::Bin;
                    if let None = self.manifest.bin {
                        return Err(InstallError::InvalidManifest(String::from("no build or bin manifest")));
                    }
                }
            }
        }

        match self.install_type {
            InstallType::Bin => self.bin()
                .map_err(|e| InstallError::BinError(e)),
            InstallType::Build => self.build()
                .map_err(|e| InstallError::BuildError(e)),
        }
    }
}
